#####
# class.py
#
# definition of 3 classes :
#     ThreeBraid
#     LinearBraid
#     LissajousBraid
#
# version info : 1.2.4  2022/10/24  H.Ogawa
#


#import math
#import cmath
#import numpy as np
#import matplotlib.pyplot as plt

from . import globalParamaters as glbl
from . import subprocedures as sub
from . import translation as tr
from . import plot as plot
from . import proc_TB as TB
from . import proc_Lin as Lin
from . import proc_Lis as Lis

from fractions import Fraction


# ThreeBraid Class

class ThreeBraid:
    def __init__(self, wd='Not Entered', silent=True, startArea='0+') -> None:
        TB.init(self)
        self.TBsw = True
        if type(wd) != str or wd != 'Not Entered':
            self.set(wd, silent=silent, startArea=startArea)

    def __str__(self):
        return('ThreeBraid defined by the sequence "{}"'.format(self.inputData))
    def clear(self) -> None:
        self.__init__()
    def help(self) -> None:
        print(TB.help())

    def set(self, dataStr, **kwargs) -> None:
        if type(dataStr) == ThreeBraid:
            self.set(dataStr.inputData)
        elif type(dataStr) == LinearBraid:
            self.set(dataStr.RawData['AB'], silent=silent, startArea=startArea)
        elif type(dataStr) == LissajousBraid:
            self.set(dataStr.lb.RawData['AB'], silent=silent, startArea=startArea)
        elif 'error' in sub.braidReprType(dataStr):
            return(sub.braidReprType(dataStr))
        else:
            TB.set(self, dataStr, **kwargs)

    def copy(self):
        return(ThreeBraid(self.inputData))

    def __add__(self, other):
        if isinstance(other, self.__class__):
            tmp = self.braid['AB'] + other.braid['AB']
            return(ThreeBraid(tmp))
        else:
            raise NotImplementedError()

    def __eq__(self, other):
        return(TB.eq(self, other))
    def __ne__(self, other):
        return(TB.ne(self, other))
    def __le__(self, other):
        return(TB.le(self, other))
    def __ge__(self, other):
        return(TB.ge(self, other))
    def __lt__(self, other):
        return(TB.lt(self, other))
    def __gt__(self, other):
        return(TB.gt(self, other))
    def ref(self, brStyle='braid'):
        return(tr.referBraidStyle(self, brStyle))

    def reverse(self):
        return(ThreeBraid(tr.ABrev(self.braid['AB']), silent=True))
    def involution(self):
        return(ThreeBraid(tr.ABinv(self.braid['AB']), silent=True))
    def reverseA(self):
        return(ThreeBraid(tr.ABrevA(self.braid['AB']), silent=True))
    def involutionA(self):
        return(ThreeBraid(tr.ABinvA(self.braid['AB']), silent=True))

    def rev(self):
        return(self.reverse(self))
    def inv(self):
        return(self.involution(self))
    def revA(self):
        return(self.reverseA(self))
    def invA(self):
        return(self.involutionA(self))

    def show(self, brStyle=[], brType=[], **kwargs) -> None:
        styles = brStyle if brStyle not in [[], '', 'all'] else glbl.braidStyles
        if type(styles) == str:
            styles = [styles]
        types = brType if brType not in [[], '', 'all'] else glbl.braidTypes
        if type(types) == str:
            types = [types]
        types = ['braidStyle', 'paramater', 'type'] + types
        print('\n'.join(sub.removeFrom([TB.brValues(self.ref(bs), types, **kwargs) for bs in styles], '')))

    def val(self, brType, brStyle='braid') -> str:
        return(self.ref(brStyle)[brType])

    if glbl.support_plot_procedures:
        def plot(self, pltType='braid', brStyle='braid', **kwargs):
            kwargs.setdefault('nonlinear', True)
            sz, sA, Ti = self.ref(brStyle)['forPlot'][:3]
            kwargs.setdefault('T', Ti)
            kwargs['startArea']=sA
            return(plot.plotTBsub(sz, pltType, **kwargs))




# LinearBraid Class

class LinearBraid(ThreeBraid):
    def __init__(self, brWord='', linear=[], lissajous=[], lissajousnew=[], startArea='0+', silent=True) -> None:
        super().__init__()
        Lin.init(self)
        if type(brWord) != str or brWord != '' or linear != [] or lissajous != [] or lissajousnew != []:
            self.set(brWord=brWord, linear=linear, lissajous=lissajous, lissajousnew=lissajousnew, startArea=startArea, silent=silent)

    def __str__(self):
        if self.TBsw:
            return(super().__str__())
        else:
            return('LinearBraid with the {} {}'.format(self.RawData['inputData'][0], sub.pprint(self.RawData['inputData'][-1])))

    def clear(self) -> None:
        self.__init__()
    def help(self) -> str:
        return(Lin.help()+super().help())
    def title(self) -> str:
        return(Lin.title(self))
    def values(self, **kwargs) -> str:
        return(self.title()+Lin.values(self, **kwargs))

    def show(self, brStyle='', brType='', **kwargs) -> None:
        if self.TBsw or brStyle != '' or brType != '':
            super().show(brStyle=brStyle, brType=brType, **kwargs)
        else:
            print(self.values(**kwargs))

    def showType(self) -> None:
        print(self.title())

    def val(self, brType, brStyle='loop3'):
        return(super().val(brType, brStyle=brStyle))

    def lissajousCurve(self, t) -> complex:
        return(Lin.lissajousCurve(t, lisType=self.LissajousType))
    def lissajousLoop(self, t) -> complex:
        return(Lin.lissajousLoop(t, lisType=self.LissajousType))
    def domain(self, t) -> list:
        return(Lin.domainOnShape(self.lissajousLoop(t)))

    def set(self, *brWord, **kwargs) -> None:
        if type(brWord) == ThreeBraid:
            self.TBsw = True
            self.brSet(brWord.inputData, startArea=brWord.startArea, silent=silent)
        elif type(brWord) == LinearBraid:
            self.set(*brWord.inputArgs, silent=silent)
        elif type(brWord) == LissajousBraid:
            self.set(*brWord.lb.inputArgs, silent=silent)
        else:
            Lin.set(self, *brWord, **kwargs)

    def brSet(self, brSeq, silent=False, **kwargs) -> None:
        super().set(brSeq, silent=silent, **kwargs)

    def setAB(self, lisType, ABseq, silent=True):
        self.clear()
        self.LissajousType = lisType
        self.LinearParam = Lin.LisT_LinP(self.LissajousType)
        self.CollisionFreeQ =Lin.collisionFreeQ(self.LissajousType)
        self.startArea = Lin.startArea(self.LissajousType)
        self.RawData['AB'] = ABseq
        self.RawData['Syzygy'] = tr.toSyzygy(ABseq, startArea=self.startArea)
        self.RawData['AB3'] = sub.oneThird(self.RawData['AB'])
        self.RawData['Syzygy3'] = sub.oneThird(self.RawData['Syzygy'])
        self.RawData['AB6'] = sub.halfAB(self.RawData['AB3'])
        self.RawData['Syzygy6'] = sub.half(self.RawData['Syzygy3'])
        self.brSet(ABseq, silent=silent)
        self.proc_LinearCancellation()
        self.TBsw = False

    def proc_LinearParamater(self, silent=False) -> None:
        self.LinearParam = Lin.LisT_LinP(self.LissajousType)
        self.CollisionFreeQ = Lin.collisionFreeQ(self.LissajousType)
        self.mainProcedure(silent=silent)

    def proc_LissajousType(self, silent=False) -> None:
        self.LissajousType = Lin.LinP_LisT(self.LinearParam)
        self.CollisionFreeQ = Lin.collisionFreeQ(self.LissajousType)
        self.mainProcedure(silent=silent)

    def mainProcedure(self, silent=False) -> None:
        sA = Lin.startArea(self.LissajousType)
        self.RawData['Syzygy'] = Lin.LinearSyzygySequence(self.LinearParam)
        self.startArea = self.RawData['Syzygy'][0]+glbl.dictSgnMinus[sA[-1]]
        self.RawData['AB'] = tr.Syzygy_AB(self.RawData['Syzygy'], startArea=self.startArea[-1], circular=True) #+'A'
        self.RawData['ABex'] = tr.AB_ABex(self.RawData['AB'], startArea=self.startArea, circular=True)
        self.RawData['Syzygy3'] = sub.oneThird(self.RawData['Syzygy'])
        self.RawData['AB3'] = sub.oneThird(self.RawData['AB'])
        self.RawData['ABex3'] = sub.oneThird(self.RawData['ABex'])
        self.RawData['Syzygy6'] = sub.half(self.RawData['Syzygy3'])
        self.RawData['AB6'] = sub.halfAB(self.RawData['AB3'])
        self.RawData['ABex6'] = sub.halfAB(self.RawData['ABex3'])
        self.brSet(self.RawData['AB'], startArea=self.startArea, silent=silent)
        self.proc_LinearCancellation()

    def proc_LinearCancellation(self) -> None:
        self.LinearCancellation['slope'] = Lin.LinearCancellationSlope(self.LinearParam[0])
        self.LinearCancellation['Syzygy'] = Lin.LinearSyzygySequence([self.LinearCancellation['slope'], 0, len(self.loop['Syzygy'])]) if self.LinearCancellation['slope'] != 0 else ''
        self.LinearCancellation['Syzygy3'] = sub.oneThird(self.LinearCancellation['Syzygy'])
        if sub.compaireCircularWords(self.loop['Syzygy'], self.LinearCancellation['Syzygy'], simple=False)[1]%2 == 1:
            self.LinearCancellation['AB'] = tr.toAB(self.LinearCancellation['Syzygy'], circular=True, startArea=self.startArea[0]+('+' if self.startArea[1] == '-' else '+'))
        else:
            self.LinearCancellation['AB'] = tr.toAB(self.LinearCancellation['Syzygy'], circular=True, startArea=self.startArea)
        self.LinearCancellation['AB3'] = sub.oneThird(self.LinearCancellation['AB'])

    def shift(self, r=1):
        self.RawData['Syzygy'] = sub.shift(self.RawData['Syzygy'], r)
        self.startArea = self.RawData['Syzygy'][0]+(self.startArea[1] if r%2 else glbl.dictSgnMinus[self.startArea[1]])
        self.RawData['AB'] = tr.Syzygy_AB(self.RawData['Syzygy'], startArea=self.startArea)
        self.RawData['Syzygy3'] = sub.oneThird(self.RawData['Syzygy'])
        self.RawData['AB3'] = sub.oneThird(self.RawData['AB'])
        self.brSet(self.RawData['Syzygy'], startArea=self.startArea, silent=True)

    def copy(self):
        return(LinearBraid(*self.inputArgs))

    def __add__(self, other):
        if isinstance(other, self.__class__):
            if self.TBsw or other.TBsw:
                return(super.__add__(other))
            else:
                tmp = self.RawData['Syzygy'] + other.RawData['Syzygy']
                return(ThreeBraid(tmp, startArea = self.startArea))
        elif isinstance(self, other.__class__):
            return(other + self)
        else:
            raise NotImplementedError()

    #def __eq__(self, other):
    #    if isinstance(other, self.__class__) or isinstance(self, other.__class__):
    #        if self.TBsw or other.TBsw:
    #            return(TB.eq(self, other))
    #        else:
    #            return(sub.compaireCircularWords(self.loop['AB'], other.loop['AB']))
    #    else:
    #        raise NotImplementedError()

    #def __ne__(self, other):
    #    if isinstance(other, self.__class__) or isinstance(self, other.__class__):
    #        if self.TBsw or other.TBsw:
    #            return(TB.ne(self, other))
    #        else:
    #            return(not sub.compaireCircularWords(self.loop['AB'], other.loop['AB']))
    #    else:
    #        raise NotImplementedError()

    #def __le__(self, other):
    #    if isinstance(other, self.__class__) or isinstance(self, other.__class__):
    #        if self.TBsw or other.TBsw:
    #            return(TB.eq(self, other))
    #        else:
    #            return(sub.inCW(self.circular['AB'], other.circular['AB']))
    #    else:
    #        raise NotImplementedError()

    #def __ge__(self, other):
    #    if isinstance(other, self.__class__) or isinstance(self, other.__class__):
    #        if self.TBsw or other.TBsw:
    #            return(TB.eq(self, other))
    #        else:
    #            return(sub.inCW(other.circular['AB'], self.circular['AB']))
    #    else:
    #        raise NotImplementedError()

    #def __lt__(self, other):
    #    if isinstance(other, self.__class__) or isinstance(self, other.__class__):
    #        if self.TBsw or other.TBsw:
    #            return(TB.eq(self, other))
    #        else:
    #            return(sub.inCW(self.circular['AB'], other.circular['AB']) and (not sub.compaireCircularWords(self.circular['AB'], other.circular['AB'])))
    #    else:
    #        raise NotImplementedError()

    #def __gt__(self, other):
    #    if isinstance(other, self.__class__) or isinstance(self, other.__class__):
    #        if self.TBsw or other.TBsw:
    #            return(TB.eq(self, other))
    #        else:
    #            return(sub.inCW(other.circular['AB'], self.circular['AB']) and (not sub.compaireCircularWords(self.circular['AB'], other.circular['AB'])))
    #    else:
    #        raise NotImplementedError()

    def reverse(self):
        wkLB = self.copy()
        wkLB.brSet(tr.ABrev(self.RawData['Syzygy']), silent=True)
        return(wkLB)
    def involution(self):
        wkLB = self.copy()
        wkLB.startArea = self.startArea[0]+('-' if self.startArea[1] == '+' else '+')
        wkLB.brSet(self.RawData['Syzygy'], silent=True)
        return(wkLB)
    def reverseA(self):
        wkLB = self.copy()
        wkLB.startArea = self.startArea[0]+('-' if self.startArea[1] == '+' else '+')
        wkLB.brSet(tr.ABrev(self.RawData['Syzygy']), silent=True)
        return(wkLB)
    def involutionA(self):
        wkLB = self.copy()
        wkLB.brSet(self.RawData['Syzygy'], silent=True)
        return(wkLB)

    def rev(self):
        return(self.reverse())
    def inv(self):
        return(self.involution())
    def revA(self):
        return(self.reverseA())
    def invA(self):
        return(self.involutionA())

    if glbl.support_plot_procedures:
        def plot(self, pltType='braid', brStyle='loop3', TB=False, **kwargs):
            if TB:
                super().plot(pltType=pltType, brStyle=brStyle, **kwargs)
            else:
                if 'T' in kwargs.keys():
                    kwargs['T_interval'] = kwargs['T']
                else:
                    kwargs.setdefault('T_interval', glbl.options['T_interval3'])
                pTl = pltType.lower()
                if 'all' in pTl:
                    return(False)
                elif 'f' in pTl:
                    return(False)
                elif 'p' in pTl or 'l' in pTl:
                    if 's' in pTl:
                        plot.plot_LissajousLoopSS(self.plotData, **kwargs)
                        return(True)
                    else:
                        plot.plot_LissajousLoop(self.plotData, **kwargs)
                        return(True)
                elif 'o' in pTl:
                    plot.plot_LissajousCurve(self.plotData, **kwargs)
                    return(True)
                elif 'b' in pTl:
                    plot.plot_LissajousBraid(self.plotData, **kwargs)
                    return(True)
                else:
                    plot.plot_LissajousCurve(self.plotData, **kwargs)
                    return(True)



                
# LissajousBraid Class

class LissajousBraid():
    def __init__(self, lisType=[], silent=False) -> None:
        Lis.init(self)
        self.lb = LinearBraid()
        if lisType != []:
            self.set(lisType, silent=silent)

    def __str__(self):
        return('LissajousBraid with the Lissajous type (new) {}'.format(sub.pprint(self.inputArgs[-1])))

    def clear(self) -> None:
        self.__init__()
    #def help(self) -> None:
    #    Lis.help()

    def show(self, *args, **kwargs) -> None:
        if len(args) > 0 and args[0] == 'KNO':
            print(Lis.KNO(self))
        else:
            print(Lis.showResults(self, *args, **kwargs))

    def val(self, brType, brStyle='loop3'):
        return(self.lb.val(brType, brStyle=brStyle))

    def set(self, lisType, **kwargs) -> None:
        if type(lisType) == ThreeBraid:
            return('error: cannot convert ThreeBraid class to LissajousBraid class.')
        elif type(lisType) == (LinearBraid, LissajousBraid):
            self.set(self.inputArgs[-1], silent=silent)
        else:
            Lis.set(self, lisType, **kwargs)
            self.braid = self.lb.braid
            self.circular = self.lb.circular
            self.loop = self.lb.loop
            self.loop3 = self.lb.loop3
            self.loop6 = self.lb.loop6
    
    def calc(self) -> None:
        Lis.calc(self)

    def __eq__(self, other):
        return(TB.eq(self, other))
    def __ne__(self, other):
        return(TB.ne(self, other))
    def __le__(self, other):
        return(TB.le(self, other))
    def __ge__(self, other):
        return(TB.ge(self, other))
    def __lt__(self, other):
        return(TB.lt(self, other))
    def __gt__(self, other):
        return(TB.gt(self, other))

    if glbl.support_plot_procedures:
        def plot(self, *args, **kwargs):
            self.lb.plot(*args, **kwargs)



