#####
# plot.py
#
# procedures for plot
#
#   plot.LissajousBraid
#   plot.ThreeBraid
#
# version info : 1.2.4  2022/10/24  H.Ogawa
#

import math
import cmath
import numpy as np
from fractions import Fraction

from . import globalParamaters as glbl
from . import subprocedures as sub
from . import translation as tr


try:
    from ipywidgets import interact, FloatSlider, IntSlider, Text, Button, Layout, HBox, VBox, interactive_output
    import matplotlib.pyplot as plt
    #import matplotlib.animation as animation
    #from matplotlib.widgets import Slider
    #from IPython.display import HTML
except:
    glbl_support_plot_procedures = False
else:
    glbl_support_plot_procedures = True


#####

def Lissajous(pltType='orbit', **kwargs):
    if 'T' in kwargs.keys():
        kwargs.setdefault('T_interval', kwargs['T'])
    else:
        kwargs.setdefault('T_interval', glbl.options['T_interval3'])
    #
    pTl = pltType.lower()
    if 'l' in pTl or 'p' in pTl:
        if 's' in pTl:
            LissajousLoopSS(**kwargs)
        else:
            LissajousLoop(**kwargs)
    elif 'o' in pTl:
        LissajousCurve(**kwargs)
    elif 'b' in pTl:
        LissajousBraid(**kwargs)
    else:
        LissajousCurve(**kwargs)


#####

def ThreeBraid(word, pltType='braid', connectSZ=False, **kwargs):
    if 'reduction' in kwargs.keys() and kwargs['reduction']:
        del kwargs['reduction']
        return(ThreeBraid(tr.reduceAB(tr.toAB(word)), pltType, **kwargs))
    #
    if sub.braidReprType(word) == 'Syzygy':
        syzygy = word
        if len(word)%2 == 0 and not(connectSZ):
            syzygy = word+tr.SZmTo(word, int(word[-1]))[1:]
            kwargs.setdefault('T', (0, 1/2))
    else:
        if sub.braidReprType(word) == 'Sigma':
            syzygy, sA, Ti  = tr.toClosureSigma(word, **kwargs)[:3]
        else:
            syzygy, sA, Ti  = tr.toClosureAB(word, **kwargs)[:3]
        kwargs['startArea'] = sA
        kwargs.setdefault('T', Ti)
        kwargs.setdefault('triangles', False)
        kwargs.setdefault('nonlinear', True)
    #
    return(plotTBsub(syzygy, pltType, **kwargs))


#####

def plotTBsub(syzygy, pltType, **kwargs):
    if 'T' in kwargs.keys():
        kwargs.setdefault('T_interval', kwargs['T'])
    #
    pTl = pltType.lower()
    if 'all' in pTl:
        plotAll(syzygy, **kwargs)
        return(True)
    elif 'f' in pTl:
        return(plotFexp(syzygy, **kwargs))
    elif 'p' in pTl or 'l' in pTl:
        if 's' in pTl:
            plotPathSS(syzygy, **kwargs)
            return(True)
        else:
            plotPath(syzygy, **kwargs)
            return(True)
    elif 'b' in pTl:
        plotBraid(syzygy, **kwargs)
        return(True)
    else:
        plotCurve(syzygy, **kwargs)
        return(True)


##### Sub-procedure saveFigure

def saveFigure(fig, savefile='', format='', dpi=0):
    print(savefile, format, dpi)
    if savefile != '':
        fn = savefile.split('.')
        print(fn)
        if format != '':
            if len(fb) > 1 and fn[-1] != format:
                fn += format
            f = '.'.join(fn)
            if dpi > 0:
                fig.savefig(f, format=format, dpi=dpi)
            else:
                fig.savefig(f, format=format)
        else:
            if dpi > 0:
                fig.savefig(savefile, dpi=dpi)
            else:
                fig.savefig(savefile)
        return(True)
    else:
        return(False)

def timeToFrame(j, steps, t0, t1):
    return(sub.inInterval(math.floor(steps*(j-t0)/(t1-t0)), (0, steps-1)))

def snapToTime(sp, t0, t1):
    if type(sp)==bool:
        return(t0)
    else:
        return(min(max(float(sp), t0), t1))


# draw Lissajous Loop

def LissajousLoop(new=True, triangles=True, linewidth=1, message='', aspect='equal', **kwargs):
    def plotLocalProcedure(m, n, d, j):
        msg.value = message
        if plotData['LissajousType'][plotData['LissajousType']['style']] != [m, n, d]:
            plotData['LissajousType'][plotData['LissajousType']['style']] = [m, n, d]
            LB_setPlotData(plotData, **opts)
        #
        psi = plotData['psi']
        #
        fig = plt.figure(figsize=opts['figsize'])
        ax = fig.add_subplot(111)
        ax.set_xlim((-opts['pathLims'], opts['pathLims']))
        ax.set_ylim((-opts['pathLims'], opts['pathLims']))
        ax.set_aspect(aspect)
        #ax.axis("off")
        #
        # 赤道
        t = np.linspace(0, 2*np.pi, 100)
        ax.plot(np.cos(t), np.sin(t), color=opts['color'][2][1])
        t = np.linspace(0, 4*np.pi/3, 3)
        ax.scatter(np.cos(t), np.sin(t), s=49, color=opts['color'][2][2])
        #
        ax.plot(*psi, color=opts['color'][2][0], linewidth=linewidth)
        #
        if triangles:
            j0 = timeToFrame(j, plotData['steps'], t0, t1)
            ax.plot(psi[0][j0], psi[1][j0], color=plotData['trcolor'][j0], marker='o', linewidth=linewidth+2)
        #
        fig.canvas.draw_idle()
        return(fig)
    #
    opts = glbl.margeOptions(kwargs)
    t0, t1 = opts['T_interval']
    plotData = {'set':False, 'steps':0, 'LissajousType':{}, 'delta':[], 't':[], 'triangles':[], 'edges':[], 'trcolor':[]}
    if new:
        plotData['LissajousType'] = {'style':'new', 'new':[4,-5,0], 'old':[]}
    else:
        plotData['LissajousType'] = {'style':'old', 'new':[], 'old':[4,-5,0]}
    #
    LB_setPlotData(plotData, **opts)
    #
    # sliders
    sliderM = IntSlider(value=4, min=-101, max=102, step=1, description='m : ')
    sliderN = IntSlider(value=-5, min=-101, max=102, step=1, description='n : ')
    #
    button = Button(description='保存', layout=Layout(width='42pt'))
    text = Text(description='ファイル', layout=Layout(width='170pt'))
    msg = Text(description='message', layout=Layout(width='215pt'))
    #
    def saveButton(b):
        fig = plotLocalProcedure(sliderM.value, sliderN.value, sliderD.value, sliderTr.value)
        if saveFigure(fig, savefile=text.value):
            msg.value = 'saved to '+text.value
        else:
            msg.value = 'error: cannot save to '+text.value
    #
    button.on_click(saveButton)
    #
    if new:
        sliderD = FloatSlider(value=0, min=-np.pi, max=3*np.pi, step=4*np.pi/200, description='phase : ')
    else:
        sliderD = FloatSlider(value=0, min=-0.5, max=1.5, step=3/200, description='phase : ')
    #
    if triangles:
        sliderTr = FloatSlider(value=t0, min=t0, max=t1, step=(t1-t0)/200, description='３動点 : ')
        uiAll = VBox([sliderM, sliderN, sliderD, sliderTr, HBox([text, button]), msg])
    else:
        sliderTr = FloatSlider(value=t0, min=t0, max=t0, step=1, description='３動点 : ')
        uiAll = VBox([sliderM, sliderN, sliderD, HBox([text, button]), msg])
    #
    out = interactive_output(plotLocalProcedure, {'m': sliderM, 'n': sliderN, 'd': sliderD, 'j': sliderTr})
    display(HBox([out, uiAll]))


# draw Lissajous Loop on Shape sphere

def LissajousLoopSS(new=True, triangles=True, linewidth=1, figsize=glbl.options['figsize'], message='', **kwargs):
    def plotLocalProcedure(m, n, d, j, elev, azim):
        msg.value = message
        if plotData['LissajousType'][plotData['LissajousType']['style']] != [m, n, d]:
            plotData['LissajousType'][plotData['LissajousType']['style']] = [m, n, d]
            LB_setPlotData(plotData, **kwargs)
        #
        xyz = plotData['lisLoop']
        #
        fig = plt.figure(figsize=figsize)
        ax = fig.add_subplot(111, projection='3d')
        #ax.set_xlim((-1,1))
        #ax.set_ylim((-1,1))
        #ax.set_zlim((-1,1))
        ax.set_box_aspect([1,1,1])
        ax.axis("off")
        #
        # 赤道
        t = np.linspace(0, 2*np.pi, 100)
        ax.plot(np.cos(t), np.sin(t), 0, color=opts['color'][2][1])
        t = np.linspace(0, 4*np.pi/3, 3)
        ax.scatter(np.cos(t), np.sin(t), 0, s=49, color=(opts['color'][2][1], opts['color'][2][2], opts['color'][2][2]))
        ax.scatter([0,0], [0,0], [1,-1], s=36, color=opts['color'][2][3])
        # loop
        ax.plot(*xyz, color=opts['color'][2][0], linewidth=linewidth)
        #
        if triangles:
            j0 = timeToFrame(j, plotData['steps'], t0, t1)
            ax.plot(xyz[0][j0], xyz[1][j0], xyz[2][j0], color=plotData['trcolor'][j0], marker='o', linewidth=linewidth+2)
        #
        ax.view_init(elev=elev, azim=azim)
        fig.canvas.draw_idle()    
        return(fig)
    #
    opts = glbl.margeOptions(kwargs)
    t0, t1 = opts['T_interval']
    plotData = {'set':False, 'steps':0, 'LissajousType':{}, 'delta':[], 't':[], 'triangles':[], 'edges':[], 'trcolor':[]}
    if new:
        plotData['LissajousType'] = {'style':'new', 'new':[4,-5,0], 'old':[]}
    else:
        plotData['LissajousType'] = {'style':'old', 'new':[], 'old':[4,-5,0]}
    #
    LB_setPlotData(plotData, **kwargs)
    #
    # sliders
    sliderELEV = IntSlider(value=opts['viewSS'][1], min=-101, max=102, step=1, description='仰角 : ')
    sliderAZIM = IntSlider(value=opts['viewSS'][0], min=-182, max=271, step=1, description='視線 : ')
    sliderM = IntSlider(value=4, min=-101, max=102, step=1, description='m : ')
    sliderN = IntSlider(value=-5, min=-101, max=102, step=1, description='n : ')
    #
    button = Button(description='保存', layout=Layout(width='42pt'))
    text = Text(description='ファイル', layout=Layout(width='170pt'))
    msg = Text(description='message', layout=Layout(width='215pt'))
    #
    def saveButton(b):
        fig = plotLocalProcedure(sliderM.value, sliderN.value, sliderD.value, sliderTr.value, sliderELEV.value, sliderAZIM.value)
        if saveFigure(fig, savefile=text.value):
            msg.value = 'saved to '+text.value
        else:
            msg.value = 'error: cannot save to '+text.value
    #
    button.on_click(saveButton)
    #
    if new:
        sliderD = FloatSlider(value=0, min=-np.pi, max=3*np.pi, step=0.002, description='phase : ')
    else:
        sliderD = FloatSlider(value=0, min=-0.5, max=1.5, step=0.001, description='phase : ')
    #
    if triangles:
        sliderTr = FloatSlider(value=t0, min=t0, max=t1, step=(t1-t0)/200, description='３動点 : ')
        uiAll = VBox([sliderM, sliderN, sliderD, sliderTr, sliderELEV, sliderAZIM, HBox([text, button]), msg])
    else:
        sliderTr = FloatSlider(value=t0, min=t0, max=t0, step=1, description='３動点 : ')
        uiAll = VBox([sliderM, sliderN, sliderD, sliderELEV, sliderAZIM, HBox([text, button]), msg])
    #
    out = interactive_output(plotLocalProcedure, {'m': sliderM, 'n': sliderN, 'd': sliderD, 'j': sliderTr, 'elev':sliderELEV, 'azim':sliderAZIM})
    display(HBox([out, uiAll]))


# draw Lissajous Curve

def LissajousCurve(new=True, triangles=True, message='', aspect='auto', **kwargs):
    def plotLocalProcedure(m, n, d, j):
        msg.value = message
        if plotData['LissajousType'][plotData['LissajousType']['style']] != [m, n, d]:
            plotData['LissajousType'][plotData['LissajousType']['style']] = [m, n, d]
            LB_setPlotData(plotData, **opts)
        #
        fig = plt.figure(figsize=opts['figsize'])
        ax = fig.add_subplot(111)
        #ax.set_xlim([-1.1,1.1])
        #ax.set_ylim([-1.1,1.1])
        ax.set_aspect(aspect)
        ax.axis("off")
        #
        ax.plot(*plotData['a'], lw=ls, color=opts['color'][0][0])
        ax.plot(*plotData['b'], lw=ls, color=opts['color'][0][1])
        ax.plot(*plotData['c'], lw=ls, color=opts['color'][0][2])
        ax.plot(plotData['a'][0][0], plotData['a'][1][0], ms=ps, color=opts['color'][0][0], marker='o')
        ax.plot(plotData['b'][0][0], plotData['b'][1][0], ms=ps, color=opts['color'][0][1], marker='o')
        ax.plot(plotData['c'][0][0], plotData['c'][1][0], ms=ps, color=opts['color'][0][2], marker='o')
        if triangles:
            j0 = timeToFrame(j, plotData['steps'], t0, t1)
            ax.plot(*plotData['edges'][j0], lw=ts, color=plotData['trcolor'][j0], alpha=0.5)
        #
        fig.canvas.draw_idle()    
        return(fig)
    #
    opts = glbl.margeOptions(kwargs)
    t0, t1 = opts['T_interval']
    opts['color'] = [x if len(x) == len(y) else y for x,y in zip(opts['color'], glbl.options['color'])]
    plotData = {'set':False, 'steps':0, 'LissajousType':{}, 'delta':[], 't':[], 'triangles':[], 'edges':[], 'trcolor':[]}
    if new:
        plotData['LissajousType'] = {'style':'new', 'new':[4,-5,0], 'old':[]}
    else:
        plotData['LissajousType'] = {'style':'old', 'new':[], 'old':[4,-5,0]}
    #
    LB_setPlotData(plotData, **opts)
    #
    ls = opts['linewidth']
    ts = ls+2
    ps = 2*ls
    #
    s = LB_setPlotSize(plotData['steps'], ls)
    r = len(plotData['edgesFill'][0][0])
    c_tr = np.concatenate([plotData['color'], [plotData['trcolor'][0]]*r])
    s_tr = np.concatenate([s, [ts]*r])
    #
    # sliders
    sliderM = IntSlider(value=4, min=-101, max=102, step=1, description='m : ')
    sliderN = IntSlider(value=-5, min=-101, max=102, step=1, description='n : ')
    #
    button = Button(description='保存', layout=Layout(width='42pt'))
    text = Text(description='ファイル', layout=Layout(width='170pt'))
    msg = Text(description='message', layout=Layout(width='215pt'))
    #
    def saveButton(b):
        fig = plotLocalProcedure(sliderM.value, sliderN.value, sliderD.value, sliderTr.value)
        if saveFigure(fig, savefile=text.value):
            msg.value = 'saved to '+text.value
        else:
            msg.value = 'error: cannot save to '+text.value
    #
    button.on_click(saveButton)
    #
    if new:
        sliderD = FloatSlider(value=0, min=-np.pi, max=3*np.pi, step=0.002, description='phase : ')
    else:
        sliderD = FloatSlider(value=0, min=-0.5, max=1.5, step=0.001, description='phase : ')
    #
    if triangles:
        sliderTr = FloatSlider(value=t0, min=t0, max=t1, step=(t1-t0)/200, description='３動点 : ')
        uiAll = VBox([sliderM, sliderN, sliderD, sliderTr, HBox([text, button]), msg])
    else:
        sliderTr = FloatSlider(value=t0, min=t0, max=t0, step=1, description='３動点 : ')
        uiAll = VBox([sliderM, sliderN, sliderD, HBox([text, button]), msg])
    #
    out = interactive_output(plotLocalProcedure, {'m': sliderM, 'n': sliderN, 'd': sliderD, 'j': sliderTr})
    display(HBox([out, uiAll]))


# draw Lissajous Braid

def LissajousBraid(new=True, triangles=True, aspect=[2,2,3], message='', **kwargs):
    def plotLocalProcedure(m, n, d, j, elev, azim):
        msg.value = message
        if plotData['LissajousType'][plotData['LissajousType']['style']] != [m, n, d]:
            plotData['LissajousType'][plotData['LissajousType']['style']] = [m, n, d]
            LB_setPlotData(plotData, **opts)
        #
        r = len(plotData['edgesFill'][0][0])
        s = LB_setPlotSize(plotData['steps'], ls)
        if triangles:
            j0 = timeToFrame(j, plotData['steps'], t0, t1)
            x_tr = np.concatenate([plotData['x'], plotData['edgesFill'][j0][0]])
            y_tr = np.concatenate([plotData['y'], plotData['edgesFill'][j0][1]])
            z_tr = np.concatenate([plotData['z'], [plotData['t'][j0]]*r])
            c_tr = np.concatenate([plotData['color'], [plotData['trcolor'][j0]]*r])
            s_tr = np.concatenate([s, [ts]*r])
        #
        fig = plt.figure(figsize=opts['figsize'])
        ax = fig.add_subplot(111, projection='3d')
        #ax.set_xlim([-1,1])
        #ax.set_ylim([-1,1])
        #ax.set_zlim([0,1/3])
        ax.set_box_aspect(aspect)
        ax.axis("off")
        #
        if triangles:
            ax.scatter(x_tr, y_tr, -z_tr, c=c_tr, s=s_tr)
        else:
            ax.scatter(plotData['x'], plotData['y'], -plotData['z'], c=plotData['color'], s=s)
        #
        ax.view_init(elev=elev, azim=azim)
        fig.canvas.draw_idle()    
        return(fig)
    #
    opts = glbl.margeOptions(kwargs)
    t0, t1 = opts['T_interval']
    plotData = {'set':False, 'steps':0, 'LissajousType':{}, 'delta':[], 't':[], 'triangles':[], 'edges':[], 'trcolor':[]}
    if new:
        plotData['LissajousType'] = {'style':'new', 'new':[4,-5,0], 'old':[]}
    else:
        plotData['LissajousType'] = {'style':'old', 'new':[], 'old':[4,-5,0]}
    #
    LB_setPlotData(plotData, **opts)
    #
    ls = opts['linewidth']
    ts = (ls+2)**2
    ps = 2*ls
    #
    s = LB_setPlotSize(plotData['steps'], ls)
    r = len(plotData['edgesFill'][0][0])
    c_tr = np.concatenate([plotData['color'], [plotData['trcolor'][0]]*r])
    s_tr = np.concatenate([s, [ts]*r])
    #
    # sliders
    sliderELEV = IntSlider(value=opts['view'][1], min=-101, max=102, step=1, description='仰角 : ')
    sliderAZIM = IntSlider(value=opts['view'][0], min=-182, max=271, step=1, description='視線 : ')
    sliderM = IntSlider(value=4, min=-101, max=102, step=1, description='m : ')
    sliderN = IntSlider(value=-5, min=-101, max=102, step=1, description='n : ')
    #
    button = Button(description='保存', layout=Layout(width='42pt'))
    text = Text(description='ファイル', layout=Layout(width='170pt'))
    msg = Text(description='message', layout=Layout(width='215pt'))
    #
    def saveButton(b):
        fig = plotLocalProcedure(sliderM.value, sliderN.value, sliderD.value, sliderTr.value, sliderELEV.value, sliderAZIM.value)
        if saveFigure(fig, savefile=text.value):
            msg.value = 'saved to '+text.value
        else:
            msg.value = 'error: cannot save to '+text.value
    #
    button.on_click(saveButton)
    #
    if new:
        sliderD = FloatSlider(value=0, min=-np.pi, max=3*np.pi, step=0.02, description='phase : ')
    else:
        sliderD = FloatSlider(value=0, min=-0.5, max=1.5, step=0.01, description='phase : ')
    #
    if triangles:
        sliderTr = FloatSlider(value=t0, min=t0, max=t1, step=(t1-t0)/200, description='３動点 : ')
        uiAll = VBox([sliderM, sliderN, sliderD, sliderTr, sliderELEV, sliderAZIM, HBox([text, button]), msg])
    else:
        sliderTr = FloatSlider(value=t0, min=t0, max=t0, step=1, description='３動点 : ')
        uiAll = VBox([sliderM, sliderN, sliderD, sliderELEV, sliderAZIM, HBox([text, button]), msg])
    #
    out = interactive_output(plotLocalProcedure, {'m': sliderM, 'n': sliderN, 'd': sliderD, 'j': sliderTr, 'elev':sliderELEV, 'azim':sliderAZIM})
    display(HBox([out, uiAll]))


# Fourier Expansion of syzygy sequence

def fourierCoefficient(data, k):
    r = len(data)
    theta = np.pi*k/r
    return(sub.chop(sum([data[s] * (1/theta**2*np.sin(theta)**2/(r/2)*np.sin(2*theta*s)) for s in range(r)])), 
           sub.chop(sum([data[s] * (1/theta**2*np.sin(theta)**2/(r/2)*np.cos(2*theta*s)) for s in range(r)])))


def fourierExpansion(data, epsilon=0.4, maxcoeff=20):
    r = len(data)-1
    alpha = Fraction(data[-1]-data[0], r)
    beta = Fraction(sum(data[:-1]), r) - Fraction(alpha*(r-1), 2) + Fraction(1, 2)
    data0 = [data[s] - (alpha*s + beta - Fraction(1, 2)) for s in range(r)]
    k = 0
    err = float(max([abs(x) for x in data0]))
    alpha0 = (alpha%3)+3
    beta0 = beta%3
    if beta0 <= Fraction(1, 2): beta0 += 3
    r0 = 2*(alpha0.denominator)
    if r0%3 != 0: r0 *= 3
    delta = Fraction(1-2*beta0, 4*alpha0)
    n = -Fraction(r0*alpha0, 6)
    m = Fraction(r0, 2) + n
    ans = [(alpha, beta, err, (m,n,delta,r0,r))]
    if err < epsilon: return(ans)
    stop = maxcoeff
    jtm = ans
    while stop > 0 and err >= epsilon:
        k += 1
        stop -= 1
        jtm = fourierCoefficient(data0, k)
        data0 = [data0[s] - jtm[0]*np.sin(2*np.pi*k/r*s) - jtm[1]*np.cos(2*np.pi*k/r*s) for s in range(r)]
        err = max([abs(x) for x in data0])
        ans += [list(jtm) + [err]]
    return(ans)


def fourierExp(syzygy, epsilon=0.4, maxcoeff=100, nonlinear=False, debug=False, **kwargs):
    if debug: print('Fexp :', syzygy, epsilon, maxcoeff, nonlinear, kwargs)
    opts = glbl.margeOptions(kwargs)
    sA = opts['startArea']
    if type(syzygy) not in (str, list):
        sz = [int(sA[0])]*2
    elif len(syzygy) == 0:
        sz = [int(sA[0])]*2
    elif len(syzygy) == 1:
        sz = [int(syzygy[0])]*2
    else:
        sz = tr.lift(tr.strToSyzygy(syzygy), type='red')
    #
    if nonlinear:
        fexp =[]
    else:
        fexp = fourierExpansion(tr.lift(sz, type='inc'), maxcoeff=1)
    if len(fexp) == 1:
        return(fexp)
    else:
        return(fourierExpansion(tr.lift(sz, type='nb'), epsilon=epsilon, maxcoeff=maxcoeff))


# values of Fourier expansion

def valFE(fexp, t, slopeadjust=0, **kwargs):
    tarr = np.array(t)
    ft = (fexp[0][0]+slopeadjust)*fexp[0][-1][-1]*tarr+fexp[0][1]
    for k in range(1, len(fexp)):
        ft += fexp[k][0]*np.sin(2*np.pi*k*tarr) + fexp[k][1]*np.cos(2*np.pi*k*tarr)
    if 'array' in str(type(ft)):
        ft = ft.astype('float')
    return(ft)


# data for plot

def braidPlotData(fexp, steps=20, **kwargs):
    opts = glbl.margeOptions(kwargs)
    r = fexp[0][-1][-1]
    t = np.linspace(*opts['T_interval'], r*steps+1)
    # 始点の向き
    if opts['startArea'][-1] == '+':
        R = np.sin(np.pi*r*t)+np.sqrt(1+np.sin(np.pi*r*t)**2)
        R = (2+np.sin(np.pi*r*t))/(2-np.sin(np.pi*r*t))
        Rs = (1+glbl.SS_range*np.sin(np.pi*r*t))/(1-glbl.SS_range*np.sin(np.pi*r*t))
        R0 = -1-glbl.SS_range*np.sin(np.pi*r*t)
        R1 = -1+glbl.SS_range*np.sin(np.pi*r*t)
    else:
        #R = -np.sin(np.pi*r*t)+np.sqrt(1+np.sin(np.pi*r*t)**2)
        R = (2-np.sin(np.pi*r*t))/(2+np.sin(np.pi*r*t))
        Rs = (1-glbl.SS_range*np.sin(np.pi*r*t))/(1+glbl.SS_range*np.sin(np.pi*r*t))
        R0 = -1+glbl.SS_range*np.sin(np.pi*r*t)
        R1 = -1-glbl.SS_range*np.sin(np.pi*r*t)
    ##Rs = np.sqrt((1+glbl.SS_range*np.sin(np.pi*r*t))/(1-glbl.SS_range*np.sin(np.pi*r*t)))
    ##Rs = 1/(1-glbl.SS_range*np.sin(np.pi*r*t))
    ##Rz = (Rs-1)/(Rs+1)
    #R0 = -2*Rs/(1+Rs) #1 + Rz
    #R1 = -2/(1+Rs) #1 - Rz
    ## 註：plotCurve と plotBraid の向きを揃えるため、上記２式に符号を加えた
    ft = valFE(fexp, t, **kwargs)
    phi = (np.cos(2*np.pi/3*ft.astype('float')), np.sin(2*np.pi/3*ft.astype('float')))
    eta0x, eta0y = [R0 * np.cos(np.pi/3*ft),  R0 * np.sin(np.pi/3*ft)]
    eta1x, eta1y = [R1 * np.cos(np.pi/3*ft), -R1 * np.sin(np.pi/3*ft)]
    rt3 = np.sqrt(3)
    a = [eta0x + eta1x, eta0y + eta1y]
    b = [(-eta0x +rt3*eta0y -eta1x -rt3*eta1y)/2, 
         (-rt3*eta0x -eta0y +rt3*eta1x -eta1y)/2]
    c = [(-eta0x -rt3*eta0y -eta1x +rt3*eta1y)/2, 
         (+rt3*eta0x -eta0y -rt3*eta1x -eta1y)/2]
    z = -(-R0**2+R1**2)/(R0**2+R1**2)
    x = np.sqrt(1 - z**2)*phi[0]
    y = np.sqrt(1 - z**2)*phi[1]
    return(t, ft, (a,b,c), ((R, R0,R1), phi), ((eta0x,eta0y), (eta1x,eta1y)), (x,y,z))


# set plot data to plotData variable

def TB_setPlotData(plotData, fexp, steps=100, **kwargs):
    opts = glbl.margeOptions(kwargs)
    #
    bPD = braidPlotData(fexp, steps=steps, **opts)
    #
    plotData['steps'] = len(bPD[0])-1
    plotData['t'] = bPD[0]
    plotData['a'], plotData['b'], plotData['c'] = bPD[2]
    plotData['x'] = np.concatenate([plotData['a'][0], plotData['b'][0], plotData['c'][0]])
    plotData['y'] = np.concatenate([plotData['a'][1], plotData['b'][1], plotData['c'][1]])
    plotData['z'] = np.concatenate([plotData['t'], plotData['t'], plotData['t']])
    #
    plotData['color'] = [opts['color'][0][j] for j in range(3) for k in range(len(plotData['t']))]
    plotData['size'] = [36 for j in range(3) for k in range(len(plotData['t'])+1)]
    plotData['delta'] = tuple(zip(*[zip(*plotData[v]) for v in ['a', 'b', 'c', 'a']]))
    plotData['edges'] = tuple(zip(*[zip(*[plotData[v][j] for v in ['a', 'b', 'c', 'a']]) for j in range(2)]))
    plotData['edgesFill'] = [[sum([list(x)[:-1] for x in np.linspace(edg[j][0:3], edg[j][1:4], 21).T], []) for j in range(2)] for edg in plotData['edges']]
    #
    plotData['trcolor'] = [opts['color'][1][1+tr.orientation(delta[0:3])] for delta in plotData['delta']]


# draw three braids

def plotBraid(word, steps=20, loop=False, aspect=[2,2,3], **kwargs):
    opts = glbl.margeOptions(kwargs)
    #fexp = fourierExp(tr.strToSyzygy(word + (word[0] if loop else '')), **opts)
    fexp = fourierExp(tr.strToSyzygy(word), **opts)
    plotData = {}
    TB_setPlotData(plotData, fexp, steps=5*steps, **opts)
    plotThreeBraid(plotData, aspect=aspect, **opts)


# draw orbits of verteces

def plotCurve(word, steps=20, loop=False, **kwargs):
    opts = glbl.margeOptions(kwargs)
    #fexp = fourierExp(tr.strToSyzygy(word + (word[0] if loop else '')), **opts)
    fexp = fourierExp(tr.strToSyzygy(word), **opts)
    plotData = {}
    TB_setPlotData(plotData, fexp, steps=5*steps, **opts)
    plotOrbit(plotData, **opts)


# draw path on C

def plotPath(word, steps=20, loop=False, triangles=True, linewidth=1, message='', snapshot=False, **kwargs):
    def plotLocalProcedure(j):
        msg.value = message
        fig = plt.figure(figsize=opts['figsize'])
        axP = fig.add_subplot(1,1,1)
        axP.set_xlim((-opts['pathLims'], opts['pathLims']))
        axP.set_ylim((-opts['pathLims'], opts['pathLims']))
        axP.set_aspect('equal')
        #axP.axis("off")
        # 赤道
        t = np.linspace(0, 2*np.pi, 100)
        axP.plot(np.cos(t), np.sin(t), color=opts['color'][2][1])
        t = np.linspace(0, 4*np.pi/3, 3)
        axP.scatter(np.cos(t), np.sin(t), s=49, color=opts['color'][2][2])
        # loop
        axP.plot(*(RPhi[0][0]*RPhi[1]), color=opts['color'][2][0], linewidth=linewidth)
        #
        if triangles:
            j0 = timeToFrame(j, s, t0, t1)
            axP.plot(RPhi[0][0][j0]*RPhi[1][0][j0], RPhi[0][0][j0]*RPhi[1][1][j0], color=opts['color'][1][2], marker='o')
        #
        fig.canvas.draw_idle()    
        return(fig)
    #
    #
    opts = glbl.margeOptions(kwargs)
    t0, t1 = opts['T_interval']
    #fexp = fourierExp(tr.strToSyzygy(word + (word[0] if loop else '')), **opts)
    fexp = fourierExp(tr.strToSyzygy(word), **opts)
    RPhi = braidPlotData(fexp, steps=5*steps, **opts)[3]
    s = len(RPhi[0][0])
    #
    # sliders
    button = Button(description='保存', layout=Layout(width='42pt'))
    text = Text(description='ファイル', layout=Layout(width='170pt'))
    msg = Text(description='message', layout=Layout(width='215pt'))
    #
    def saveButton(b):
        fig = plotLocalProcedure(sliderTr.value)
        if saveFigure(fig, savefile=text.value):
            msg.value = 'saved to '+text.value
        else:
            msg.value = 'error: cannot save'
    #
    button.on_click(saveButton)
    #
    if type(snapshot) == bool and snapshot == False:
        if triangles:
            sliderTr = FloatSlider(value=t0, min=t0, max=t1, step=10*(t1-t0)/s, description='３動点 : ')
            uiAll = VBox([sliderTr, HBox([text, button]), msg])
        else:
            sliderTr = FloatSlider(value=t0, min=t0, max=t0, step=1, description='３動点 : ')
            uiAll = VBox([HBox([text, button]), msg])
        #
        out = interactive_output(plotLocalProcedure, {'j': sliderTr})
        display(HBox([out, uiAll]))
        #
    else:
        triangles = True
        plotLocalProcedure(snapToTime(snapshot, t0, t1))
        plt.show()


# draw path on Shape sphere

def plotPathSS(word, triangles=True, steps=20, loop=False, linewidth=1, message='', snapshot=False, **kwargs):
    def plotLocalProcedure(j, elev, azim):
        msg.value = message
        fig = plt.figure(figsize=opts['figsize'])
        axSS = fig.add_subplot(1,1,1, projection='3d')
        #axSS.set_xlim([-1,1])
        #axSS.set_ylim([-1,1])
        #axSS.set_zlim([-1,1])
        axSS.set_box_aspect([1,1,1])
        axSS.axis("off")
        # 赤道
        t = np.linspace(0, 2*np.pi, 100)
        axSS.plot(np.cos(t), np.sin(t), 0, color=opts['color'][2][1])
        t = np.linspace(0, 4*np.pi/3, 3)
        axSS.scatter(np.cos(t), np.sin(t), 0, s=49, color=(opts['color'][2][1], opts['color'][2][2], opts['color'][2][2]))
        axSS.scatter([0,0], [0,0], [1,-1], s=36, color=opts['color'][2][3])
        # loop
        axSS.plot(*xyz, color=opts['color'][2][0], linewidth=linewidth)
        #
        if triangles:
            j0 = timeToFrame(j, s, t0, t1)
            axSS.plot(xyz[0][j0], xyz[1][j0], xyz[2][j0], color=opts['color'][1][2], marker='o')
        #
        axSS.view_init(elev=elev, azim=azim)
        fig.canvas.draw_idle()    
        return(fig)
    #
    #
    opts = glbl.margeOptions(kwargs)
    t0, t1 = opts['T_interval']
    #fexp = fourierExp(tr.strToSyzygy(word + (word[0] if loop else '')), **opts)
    fexp = fourierExp(tr.strToSyzygy(word), **opts)
    xyz = braidPlotData(fexp, steps=5*steps, **opts)[-1]
    s = len(xyz[0])
    #
    # sliders
    sliderELEV = IntSlider(value=opts['viewSS'][1], min=-101, max=102, step=1, description='仰角 : ')
    sliderAZIM = IntSlider(value=opts['viewSS'][0], min=-182, max=271, step=1, description='視線 : ')
    #
    button = Button(description='保存', layout=Layout(width='42pt'))
    text = Text(description='ファイル', layout=Layout(width='170pt'))
    msg = Text(description='message', layout=Layout(width='215pt'))
    #
    def saveButton(b):
        fig = plotLocalProcedure(sliderTr.value, sliderELEV.value, sliderAZIM.value)
        if saveFigure(fig, savefile=text.value):
            msg.value = 'saved to '+text.value
        else:
            msg.value = 'error: cannot save to '+text.value
    #
    button.on_click(saveButton)
    #
    if type(snapshot) == bool and snapshot == False:
        if triangles:
            sliderTr = FloatSlider(value=t0, min=t0, max=t1, step=10*(t1-t0)/s, description='３動点 : ')
            uiAll = VBox([sliderTr, sliderELEV, sliderAZIM, HBox([text, button]), msg])
        else:
            sliderTr = FloatSlider(value=t0, min=t0, max=t0, step=1, description='t : ')
            uiAll = VBox([sliderELEV, sliderAZIM, HBox([text, button]), msg])
        #
        out = interactive_output(plotLocalProcedure, {'j': sliderTr, 'elev':sliderELEV, 'azim':sliderAZIM})
        display(HBox([out, uiAll]))
        #
    else:
        triangles = True
        plotLocalProcedure(snapToTime(snapshot, t0, t1), opts['viewSS'][1], opts['viewSS'][0])
        plt.show()


# plot Fourier expansion

def plotFexp(word, steps=20, loop=False, slopeadjust=0, linewidth=1, **kwargs):
    opts = glbl.margeOptions(kwargs)
    #syzygy = tr.strToSyzygy(word + (word[0] if loop else ''))
    syzygy = tr.strToSyzygy(word)
    fexp = fourierExp(syzygy, **opts)
    r = fexp[0][-1][-1]
    syzygy0 = [s + slopeadjust*i for i,s in enumerate(tr.lift(syzygy) if len(fexp) < 2 else tr.lift(syzygy, type='nb'))]
    #
    fig = plt.figure(figsize=opts['figsize'])
    axFE = fig.add_subplot(1,1,1)
    #
    mm = [min(syzygy0), max(syzygy0)+1]
    if mm[1]-mm[0]<4:
        axFE.set_yticks([x for x in list(range(*mm))])
    else:
        axFE.set_yticks([x for x in list(range(*mm)) if (x-syzygy0[0])%3 == 0])
    axFE.set_xticks([x for x in [k*(int(r/6+1/2)) for k in range(0, 7)] if 0 <= x <= r+1])
    #
    if len(fexp)>1: axFE.plot(syzygy0, color='black')
    if r < 25: axFE.scatter(range(len(syzygy0)), syzygy0, color='black')
    #
    t = np.linspace(*opts['T_interval'], r*steps+1)
    ft = valFE(fexp, t, slopeadjust=slopeadjust, **opts) - Fraction(1, 2)
    axFE.plot(r*t, ft, color='red', linewidth=linewidth)
    #
    plt.show()
    return(sub.pprintR(fexp))


def SStoC(phi):
    if not(type(phi) in (list, tuple)):
        return(False)
    elif type(phi[0]) in (list, tuple):
        eta0, eta1 = phi
    else:
        eta0, eta1 = phi, (1,0)
    eta0x, eta0y = eta0
    eta1x, eta1y = eta1
    rt3 = np.sqrt(3)
    a = (sub.chop(eta0x + eta1x), sub.chop(eta0y + eta1y))
    b = (sub.chop((-eta0x +rt3*eta0y -eta1x -rt3*eta1y)/2), 
         sub.chop((-rt3*eta0x -eta0y +rt3*eta1x -eta1y)/2))
    c = (sub.chop((-eta0x -rt3*eta0y -eta1x +rt3*eta1y)/2), 
         sub.chop((+rt3*eta0x -eta0y -rt3*eta1x -eta1y)/2))
    return((a, b, c), ((0,0), tuple(eta0), tuple(eta1)))


def CtoSS(abc):
    if not(type(abc) in (list, tuple)):
        return(False)
    else:
        a,b,c = abc
    ax, ay = a
    bx, by = b
    cx, cy = c
    rt3 = np.sqrt(3)
    eta = ((sub.chop((ax + bx + cx)/3), 
            sub.chop((ay + by + cy)/3)), 
           (sub.chop((ax + (-bx-rt3*by)/2 + (-cx+rt3*cy)/2)/3), 
            sub.chop((ay + ( rt3*bx-by)/2 + (-rt3*cx-cy)/2)/3)), 
           (sub.chop((ax + (-bx+rt3*by)/2 + (-cx-rt3*cy)/2)/3), 
            sub.chop((ay + (-rt3*bx-by)/2 + ( rt3*cx-cy)/2)/3)))
    phi = (sub.chop((eta[1][0]*eta[2][0]+eta[1][1]*eta[2][1])/(eta[2][0]**2+eta[2][1]**2)), 
           sub.chop((-eta[1][0]*eta[2][1]+eta[1][1]*eta[2][0])/(eta[2][0]**2+eta[2][1]**2)))
    return(phi, eta)


def phiBr(fexp, t):
    R0 = np.sin(np.pi*(r*t+1/4))
    R1 = -np.sin(np.pi*(r*t-1/4))
    ft = valFE(fexp, t)
    return(ft,
           ((R0 * np.cos(np.pi/3*ft),  R0 * np.sin(np.pi/3*ft)), 
            (R1 * np.cos(np.pi/3*ft), -R1 * np.sin(np.pi/3*ft))))


# plot path and braid

def plotAll(word, loop=False, steps=20, aspect=[2,2,3], **kwargs):
    opts = glbl.margeOptions(kwargs)
    #sz = tr.strToSyzygy(word + (word[0] if loop else ''))
    sz = tr.strToSyzygy(word)
    fexp = fourierExp(sz, **opts)
    bPD = braidPlotData(fexp, steps=5*steps, **opts)
    #
    t, ft, (a,b,c), RPhi = bPD[:4]
    x = np.concatenate([a[0], b[0], c[0]])
    y = np.concatenate([a[1], b[1], c[1]])
    z = np.concatenate([t, t, t])
    clr = [opts['color'][0][j] for j in range(3) for k in range(len(bPD[0]))]
    #
    ls = opts['linewidth']
    ts = (ls+2)**2
    ps = 2*ls
    s = LB_setPlotSize(len(t)-1, ls)
    #
    fig = plt.figure(figsize=opts['figsizeL'])
    axP = fig.add_subplot(1,3,1)
    axB = fig.add_subplot(1,3,2, projection='3d')
    axO = fig.add_subplot(1,3,3)
    #
    axP.set_xlim((-opts['pathLims'],opts['pathLims']))
    axP.set_ylim((-opts['pathLims'],opts['pathLims']))
    axP.set_aspect('equal')
    t = np.linspace(0, 2*np.pi, 100)
    axP.plot(np.cos(t), np.sin(t), color=opts['color'][2][1])
    t = np.linspace(0, 4*np.pi/3, 3)
    axP.scatter(np.cos(t), np.sin(t), s=49, color=opts['color'][2][2])
    axP.plot(*(RPhi[0][0]*RPhi[1]), color=opts['color'][2][0], linewidth=1)
    #
    axB.set_box_aspect(aspect)
    axB.axis("off")
    axB.view_init(elev=opts['view'][1], azim=opts['view'][0])
    axB.scatter(x, y, -z, c=clr, s=s)
    #
    axO.set_aspect('equal')
    axO.axis("off")
    axO.plot(*a, lw=ls, color=opts['color'][0][0])
    axO.plot(*b, lw=ls, color=opts['color'][0][1])
    axO.plot(*c, lw=ls, color=opts['color'][0][2])
    axO.plot(a[0][0], a[1][0], ms=ps, color=opts['color'][0][0], marker='o')
    axO.plot(b[0][0], b[1][0], ms=ps, color=opts['color'][0][1], marker='o')
    axO.plot(c[0][0], c[1][0], ms=ps, color=opts['color'][0][2], marker='o')
    #
    plt.show()
    #


# plot path and braid

def plotAll4(word, loop=False, steps=20, aspect=[2,2,3], **kwargs):
    opts = glbl.margeOptions(kwargs)
    #sz = tr.strToSyzygy(word + (word[0] if loop else ''))
    sz = tr.strToSyzygy(word)
    fexp = fourierExp(sz, **opts)
    bPD = braidPlotData(fexp, steps=5*steps, **opts)
    #
    t, ft, (a,b,c), RPhi = bPD[:4]
    #RPhi = bPD[3]
    #a, b, c = bPD[2]
    x = np.concatenate([a[0], b[0], c[0]])
    y = np.concatenate([a[1], b[1], c[1]])
    #z = np.concatenate([bPD[0], bPD[0], bPD[0]])
    z = np.concatenate([t, t, t])
    clr = [opts['color'][0][j] for j in range(3) for k in range(len(bPD[0]))]
    #
    ls = opts['linewidth']
    ts = (ls+2)**2
    ps = 2*ls
    #s = LB_setPlotSize(len(bPD[0])-1, ls)
    s = LB_setPlotSize(len(t)-1, ls)
    #
    fig = plt.figure(figsize=opts['figsizeL'])
    axF = fig.add_subplot(1,4,1)
    axP = fig.add_subplot(1,4,2)
    axB = fig.add_subplot(1,4,3, projection='3d')
    axO = fig.add_subplot(1,4,4)
    #
    axF.plot(tr.lift(sz, type=('inc' if len(fexp)==1 else 'nb')))        
    axF.plot((len(sz)-1)*t, ft-1/2)
    #axF.axis("off")
    #
    axP.set_xlim((-opts['pathLims'],opts['pathLims']))
    axP.set_ylim((-opts['pathLims'],opts['pathLims']))
    axP.set_aspect('equal')
    t = np.linspace(0, 2*np.pi, 100)
    axP.plot(np.cos(t), np.sin(t), color=opts['color'][2][1])
    t = np.linspace(0, 4*np.pi/3, 3)
    axP.scatter(np.cos(t), np.sin(t), s=49, color=opts['color'][2][2])
    axP.plot(*(RPhi[0][0]*RPhi[1]), color=opts['color'][2][0], linewidth=1)
    #
    axB.set_box_aspect(aspect)
    axB.axis("off")
    axB.view_init(elev=opts['view'][1], azim=opts['view'][0])
    axB.scatter(x, y, -z, c=clr, s=s)
    #
    axO.set_aspect('equal')
    axO.axis("off")
    axO.plot(*a, lw=ls, color=opts['color'][0][0])
    axO.plot(*b, lw=ls, color=opts['color'][0][1])
    axO.plot(*c, lw=ls, color=opts['color'][0][2])
    axO.plot(a[0][0], a[1][0], ms=ps, color=opts['color'][0][0], marker='o')
    axO.plot(b[0][0], b[1][0], ms=ps, color=opts['color'][0][1], marker='o')
    axO.plot(c[0][0], c[1][0], ms=ps, color=opts['color'][0][2], marker='o')
    #
    plt.show()
    #


# set plot data for Lissajous Braid

def LB_setPlotData(plotData, steps=0, **kwargs):
    opts = glbl.margeOptions(kwargs)
    m, n = plotData['LissajousType'][plotData['LissajousType']['style']][0:2]
    plotData['steps'] = sts = max(abs(m), abs(n))*50 if steps == 0 else steps
    d = float(plotData['LissajousType'][plotData['LissajousType']['style']][2])
    #
    plotData['t'] = np.linspace(*opts['T_interval'], sts+1)
    if plotData['LissajousType']['style'] == 'new':
        plotData['a'] = np.sin(2*np.pi*m*(plotData['t']-1/3)), np.sin(2*np.pi*n*(plotData['t']-1/3)+d)
        plotData['b'] = np.sin(2*np.pi*m*(plotData['t']    )), np.sin(2*np.pi*n* plotData['t']     +d)
        plotData['c'] = np.sin(2*np.pi*m*(plotData['t']+1/3)), np.sin(2*np.pi*n*(plotData['t']+1/3)+d)
    else:
        plotData['a'] = np.sin(2*np.pi*m*(plotData['t']-1/3+d)), np.sin(2*np.pi*n*(plotData['t']-1/3))
        plotData['b'] = np.sin(2*np.pi*m*(plotData['t']    +d)), np.sin(2*np.pi*n*(plotData['t']))
        plotData['c'] = np.sin(2*np.pi*m*(plotData['t']+1/3+d)), np.sin(2*np.pi*n*(plotData['t']+1/3))
    plotData['x'] = np.concatenate([plotData['a'][0], plotData['b'][0], plotData['c'][0]])
    plotData['y'] = np.concatenate([plotData['a'][1], plotData['b'][1], plotData['c'][1]])
    plotData['z'] = np.concatenate([plotData['t'], plotData['t'], plotData['t']])
    #
    plotData['color'] = [opts['color'][0][j] for j in range(3) for k in range(sts+1)]
    plotData['size'] = [36 for j in range(3) for k in range(sts+1)]
    plotData['delta'] = tuple(zip(*[zip(*plotData[v]) for v in ['a', 'b', 'c', 'a']]))
    plotData['edges'] = tuple(zip(*[zip(*[plotData[v][j] for v in ['a', 'b', 'c', 'a']]) for j in range(2)]))
    plotData['edgesFill'] = [[sum([list(x)[:-1] for x in np.linspace(edg[j][0:3], edg[j][1:4], 21).T], []) for j in range(2)] for edg in plotData['edges']]
    #
    # A = B = 1 -> A/B = ABrate
    wk1 = [[sub.R2toC(x, AB=glbl.options['ABrate']) for x in delta[:3]] for delta in plotData['delta']]
    #
    wk = [sub.CtoR2(tr.cPsi(x)) for x in wk1]
    plotData['psi'] = [[psi[0] for psi in wk], [psi[1] for psi in wk]]    
    plotData['psi2'] = plotData['psi']
    #
    wk = [sub.C2toSS(*tr.cPsiT(x)) for x in wk1]
    plotData['lisLoop'] = [
        [xyz[0] for xyz in wk],
        [xyz[1] for xyz in wk],
        [xyz[2] for xyz in wk]]
    #
    plotData['trcolor'] = [opts['color'][1][1+tr.orientation(delta[:3])] for delta in plotData['delta']]


def LB_setPlotSize(steps, linewidth):
    s = [linewidth**2]*(3*steps+3)
    for j in range(3):
        s[j*steps+j] = 4*linewidth**2
        s[(j+1)*steps+j] = 4*linewidth**2
    return(s)


# draw Lissajous Loop

def plot_LissajousLoop(plotData, triangles=True, linewidth=1, snapshot=False, message='', **kwargs):
    def plotLocalProcedure(j):
        msg.value = message
        fig = plt.figure(figsize=opts['figsize'])
        axP = fig.add_subplot(1,1,1)
        axP.set_xlim((-opts['pathLims'], opts['pathLims']))
        axP.set_ylim((-opts['pathLims'], opts['pathLims']))
        axP.set_aspect('equal')
        #axP.axis("off")
        # 赤道
        t = np.linspace(0, 2*np.pi, 100)
        axP.plot(np.cos(t), np.sin(t), color=opts['color'][2][1])
        t = np.linspace(0, 4*np.pi/3, 3)
        axP.scatter(np.cos(t), np.sin(t), s=49, color=opts['color'][2][2])
        # loop
        axP.plot(*psi, color=opts['color'][2][0], linewidth=linewidth)
        #
        if triangles:
            j0 = timeToFrame(j, plotData['steps'], t0, t1)
            axP.plot(psi[0][j0], psi[1][j0], color=opts['color'][1][2], marker='o', linewidth=linewidth+1)
        #
        fig.canvas.draw_idle()    
        return(fig)
    #
    #
    kwargs.setdefault('T_interval', glbl.options['T_interval3'])
    opts = glbl.margeOptions(kwargs)
    t0, t1 = opts['T_interval']
    LB_setPlotData(plotData, **opts)
    psi = plotData['psi']
    #
    # sliders
    button = Button(description='保存', layout=Layout(width='42pt'))
    text = Text(description='ファイル', layout=Layout(width='170pt'))
    msg = Text(description='message', layout=Layout(width='215pt'))
    #
    def saveButton(b):
        fig = plotLocalProcedure(sliderTr.value)
        if saveFigure(fig, savefile=text.value):
            msg.value = ' '+text.value
        else:
            msg.value = 'error: cannot save to '+text.value
    #
    button.on_click(saveButton)
    #
    if type(snapshot) == bool and snapshot == False:
        if triangles:
            sliderTr = FloatSlider(value=t0, min=t0, max=t1, step=(t1-t0)/200, description='３動点 : ')
            uiAll = VBox([sliderTr, HBox([text, button]), msg])
        else:
            sliderTr = FloatSlider(value=t0, min=t0, max=t0, step=1, description='３動点 : ')
            uiAll = VBox([HBox([text, button]), msg])
        #
        out = interactive_output(plotLocalProcedure, {'j': sliderTr})
        display(HBox([out, uiAll]))
        #
    else:
        triangles = True
        plotLocalProcedure(snapToTime(snapshot, t0, t1))
        plt.show()


# draw Lissajous Loop on Shape sphere

def plot_LissajousLoopSS(plotData, triangles=True, linewidth=1, message='', snapshot=False, **kwargs):
    def plotLocalProcedure(j, elev, azim):
        msg.value = message
        fig = plt.figure(figsize=opts['figsize'])
        axSS = fig.add_subplot(1,1,1, projection='3d')
        #axSS.set_xlim([-1,1])
        #axSS.set_ylim([-1,1])
        #axSS.set_zlim([-1,1])
        axSS.set_box_aspect([1,1,1])
        axSS.axis("off")
        # 赤道
        t = np.linspace(0, 2*np.pi, 100)
        axSS.plot(np.cos(t), np.sin(t), 0, color=opts['color'][2][1])
        t = np.linspace(0, 4*np.pi/3, 3)
        axSS.scatter(np.cos(t), np.sin(t), 0, s=49, color=(opts['color'][2][1], opts['color'][2][2], opts['color'][2][2]))
        axSS.scatter([0,0], [0,0], [1,-1], s=36, color=opts['color'][2][3])
        # loop
        axSS.plot(*xyz, color=opts['color'][2][0], linewidth=linewidth)
        #
        if triangles:
            j0 = timeToFrame(j, plotData['steps'], t0, t1)
            axSS.plot(xyz[0][j0], xyz[1][j0], xyz[2][j0], color=opts['color'][1][2], marker='o', linewidth=linewidth+2)
        #
        axSS.view_init(elev=elev, azim=azim)
        fig.canvas.draw_idle()    
        return(fig)
    #
    #
    kwargs.setdefault('T_interval', glbl.options['T_interval3'])
    opts = glbl.margeOptions(kwargs)
    t0, t1 = opts['T_interval']
    LB_setPlotData(plotData, **opts)
    xyz = plotData['lisLoop']
    #
    # sliders
    sliderELEV = IntSlider(value=opts['viewSS'][1], min=-101, max=102, step=1, description='仰角 : ')
    sliderAZIM = IntSlider(value=opts['viewSS'][0], min=-182, max=271, step=1, description='視線 : ')
    #
    button = Button(description='保存', layout=Layout(width='42pt'))
    text = Text(description='ファイル', layout=Layout(width='170pt'))
    msg = Text(description='message', layout=Layout(width='215pt'))
    #
    def saveButton(b):
        fig = plotLocalProcedure(sliderTr.value, sliderELEV.value, sliderAZIM.value)
        if saveFigure(fig, savefile=text.value):
            msg.value = 'saved to '+text.value
        else:
            msg.value = 'error: cannot save to '+text.value
    #
    button.on_click(saveButton)
    #
    #
    if type(snapshot) == bool and snapshot == False:
        if triangles:
            sliderTr = FloatSlider(value=t0, min=t0, max=t1, step=(t1-t0)/200, description='３動点 : ')
            uiAll = VBox([sliderTr, sliderELEV, sliderAZIM, HBox([text, button]), msg])
        else:
            sliderTr = FloatSlider(value=t0, min=t0, max=t0, step=1, description='３動点 : ')
            uiAll = VBox([sliderELEV, sliderAZIM, HBox([text, button]), msg])
        #
        out = interactive_output(plotLocalProcedure, {'j': sliderTr, 'elev':sliderELEV, 'azim':sliderAZIM})
        display(HBox([out, uiAll]))
        #
    else:
        triangles = True
        plotLocalProcedure(snapToTime(snapshot, t0, t1), opts['viewSS'][1], opts['viewSS'][0])
        plt.show()


# draw Lissajous Curves

def plot_LissajousCurve(plotData, **kwargs):
    opts = glbl.margeOptions(kwargs)
    LB_setPlotData(plotData, **opts)
    plotOrbit(plotData, **opts)


# Sub-procedure to draw orbits of verteces

def plotOrbit(plotData, triangles=False, steps=20, trianglesize=6, snapshot=False, message='', **kwargs):
    def plotLocalProcedure(j):
        msg.value = message
        #
        fig = plt.figure(figsize=opts['figsize'])
        ax = fig.add_subplot(111)
        #ax.set_xlim([-2.1,2.1])
        #ax.set_ylim([-2.1,2.1])
        #ax.set_aspect(aspect)
        ax.axis("off")
        #
        ax.plot(*plotData['a'], lw=ls, color=opts['color'][0][0])
        ax.plot(*plotData['b'], lw=ls, color=opts['color'][0][1])
        ax.plot(*plotData['c'], lw=ls, color=opts['color'][0][2])
        ax.plot(plotData['a'][0][0], plotData['a'][1][0], ms=ps, color=glbl.options['color'][0][0], marker='o')
        ax.plot(plotData['b'][0][0], plotData['b'][1][0], ms=ps, color=glbl.options['color'][0][1], marker='o')
        ax.plot(plotData['c'][0][0], plotData['c'][1][0], ms=ps, color=glbl.options['color'][0][2], marker='o')
        if triangles:
            j0 = timeToFrame(j, plotData['steps'], t0, t1)
            ax.plot(*plotData['edges'][j0], lw=ts, color=plotData['trcolor'][j0], alpha=0.5)
        #
        fig.canvas.draw_idle()    
        return(fig)
    #
    opts = glbl.margeOptions(kwargs)
    t0, t1 = opts['T_interval']
    ls = opts['linewidth']
    ts = ls+2
    ps = 2*ls
    #
    s = LB_setPlotSize(plotData['steps'], ls)
    r = len(plotData['edgesFill'][0][0])
    c_tr = np.concatenate([plotData['color'], [plotData['trcolor'][0]]*r])
    s_tr = np.concatenate([s, [ts]*r])
    #
    button = Button(description='保存', layout=Layout(width='42pt'))
    text = Text(description='ファイル', layout=Layout(width='170pt'))
    msg = Text(description='message', layout=Layout(width='215pt'))
    #
    def saveButton(b):
        fig = plotLocalProcedure(sliderTr.value)
        if saveFigure(fig, savefile=text.value):
            msg.value = 'saved to '+text.value
        else:
            msg.value = 'error: cannot save to '+text.value
    #
    button.on_click(saveButton)
    #
    if type(snapshot) == bool and snapshot == False:
        if triangles:
            sliderTr = FloatSlider(value=t0, min=t0, max=t1, step=(t1-t0)/200, description='３動点 : ', horizontal=False)
            uiAll = VBox([sliderTr, HBox([text, button]), msg])
        else:
            sliderTr = FloatSlider(value=t0, min=t0, max=t0, step=1, description='３動点 : ', horizontal=False)
            uiAll = VBox([HBox([text, button]), msg])
        out = interactive_output(plotLocalProcedure, {'j': sliderTr})
        display(HBox([out, uiAll]))
    else:
        triangles = True
        plotLocalProcedure(snapToTime(snapshot, t0, t1))
        plt.show()


# draw Lissajous Braid

def plot_LissajousBraid(plotData, aspect=[2,2,3], **kwargs):
    opts = glbl.margeOptions(kwargs)
    LB_setPlotData(plotData, **opts)
    plotThreeBraid(plotData, aspect=aspect, **opts)


# Sub-procedure to draw three braids

def plotThreeBraid(plotData, triangles=False, trianglesize=6, aspect=[2,2,3], snapshot=False, message='', **kwargs):
    def plotLocalProcedure(j, elev, azim):
        msg.value = message
        s = LB_setPlotSize(plotData['steps'], ls)
        if triangles:
            j0 = timeToFrame(j, plotData['steps'], t0, t1)
            x_tr = np.concatenate([plotData['x'], plotData['edgesFill'][j0][0]])
            y_tr = np.concatenate([plotData['y'], plotData['edgesFill'][j0][1]])
            z_tr = np.concatenate([plotData['z'], [plotData['t'][j0]]*r])
            c_tr = np.concatenate([plotData['color'], [plotData['trcolor'][j0]]*r])
            s_tr = np.concatenate([s, [ts]*r])
        #
        fig = plt.figure(figsize=opts['figsize'])
        ax = fig.add_subplot(111, projection='3d')
        #ax.set_xlim([-1,1])
        #ax.set_ylim([-1,1])
        #ax.set_zlim([t0,t1])
        ax.set_box_aspect(aspect)
        ax.axis("off")
        #
        ax.view_init(elev=elev, azim=azim)
        if triangles:
            ax.scatter(x_tr, y_tr, -z_tr, c=c_tr, s=s_tr)
        else:
            ax.scatter(plotData['x'], plotData['y'], -plotData['z'], c=plotData['color'], s=s)
        #
        fig.canvas.draw_idle()    
        return(fig)
    #
    opts = glbl.margeOptions(kwargs)
    t0, t1 = opts['T_interval']
    ls = opts['linewidth']
    ts = (ls+2)**2
    ps = 2*ls
    #
    s = LB_setPlotSize(plotData['steps'], ls)
    r = len(plotData['edgesFill'][0][0])
    c_tr = np.concatenate([plotData['color'], [plotData['trcolor'][0]]*r])
    s_tr = np.concatenate([s, [ts]*r])
    #
    # sliders
    sliderELEV = IntSlider(value=opts['view'][1], min=-101, max=102, step=1, description='仰角 : ')
    sliderAZIM = IntSlider(value=opts['view'][0], min=-182, max=271, step=1, description='視線 : ')
    #
    button = Button(description='保存', layout=Layout(width='42pt'))
    text = Text(description='ファイル', layout=Layout(width='170pt'))
    msg = Text(description='message', layout=Layout(width='215pt'))
    #
    def saveButton(b):
        fig = plotLocalProcedure(sliderTr.value, sliderELEV.value, sliderAZIM.value)
        if saveFigure(fig, savefile=text.value):
            msg.value = 'saved to '+text.value
        else:
            msg.value = 'error: cannot save to '+text.value
    #
    button.on_click(saveButton)
    #
    if type(snapshot) == bool and snapshot == False:
        if triangles:
            sliderTr = FloatSlider(value=t0, min=t0, max=t1, step=(t1-t0)/200, description='３動点 : ')
            uiAll = VBox([sliderTr, sliderELEV, sliderAZIM, HBox([text, button]), msg])
        else:
            sliderTr = FloatSlider(value=t0, min=t0, max=t0, step=1, description='３動点 : ')
            uiAll = VBox([sliderELEV, sliderAZIM, HBox([text, button]), msg])
        #
        out = interactive_output(plotLocalProcedure, {'j': sliderTr, 'elev':sliderELEV, 'azim':sliderAZIM})
        display(HBox([out, uiAll]))
        #
    else:
        triangles = True
        plotLocalProcedure(snapToTime(snapshot, t0, t1), opts['view'][1], opts['view'][0])
        plt.show()


